#!/usr/bin/env python
from . import errors as exceptions
import nomenclate.settings as settings
from . import tools


class TokenAttr(tools.Serializable):
    """ A TokenAttr represents a string token that we want to replace in a given nomenclate.core.formatter.FormatString
        It has 3 augmentation properties:

            TokenAttr().case
            TokenAttr().prefix
            TokenAttr().suffix

        These three settings enforce that after final rendering (finding matches in the config for the current
        token's label and any custom rendering syntax the user added) we will upper case the result and add
        either the given prefix or suffix no matter what.

    """

    SERIALIZE_ATTRS = ["token", "label", "case", "prefix", "suffix"]

    def __init__(self, token="", label=None, case="", prefix="", suffix=""):
        """

        :param label: str, the label is represents the value we want to replace the given token with
        :param token: str, the raw name for the token to be used
        """
        try:
            self.validate_entries(label)
        except exceptions.ValidationError:
            label = None
        self.validate_entries(token)

        self.raw_string = label if label is not None else ""
        self.raw_token = token
        self.case = case
        self.prefix = prefix
        self.suffix = suffix

    @property
    def token(self):
        """ Get or set the current token. Setting the token to a new value means it will be validated and then
            added as the internal "raw_token" which will be used to look up any given config value or be used if
            no config value is found.

        """
        return self.raw_token.lower()

    @token.setter
    def token(self, token):
        self.validate_entries(token)
        self.raw_token = token

    @property
    def label(self):
        """ Get or set the current token's "label" or value. Setting the label to a new value means it will be added
            as the internal "raw_string" which will be used to look up any given config value or be used if no config
            value is found.

        """
        return self.raw_string

    @label.setter
    def label(self, label):
        self.validate_entries(label)
        try:
            self.raw_string = int(label)
        except ValueError:
            self.raw_string = label

    def set(self, value):
        self.label = value

    @staticmethod
    def validate_entries(*entries):
        for entry in entries:
            if isinstance(entry, (str, int)) or entry is None:
                continue
            else:
                raise exceptions.ValidationError(
                    "Invalid type %s, expected %s" % (type(entry), str)
                )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            self_attrs = [getattr(self, attr) for attr in self.SERIALIZE_ATTRS]
            other_attrs = [getattr(other, attr) for attr in self.SERIALIZE_ATTRS]
            return self_attrs == other_attrs
        else:
            return False

    def __ne__(self, other):
        return (self.token, self.label) != (other.token, other.label)

    def __lt__(self, other):
        return (self.token, self.label) < (other.token, other.label)

    def __le__(self, other):
        return (self.token, self.label) <= (other.token, other.label)

    def __gt__(self, other):
        return (self.token, self.label) > (other.token, other.label)

    def __ge__(self, other):
        return (self.token, self.label) >= (other.token, other.label)

    def __str__(self):
        return "%r" % (self)

    def __repr__(self):
        return "<%s (%s): %r>" % (self.__class__.__name__, self.raw_token, self.to_json())

    def to_json(self):
        return {
            "token": self.raw_token,
            "label": self.raw_string,
            "case": self.case,
            "prefix": self.prefix,
            "suffix": self.suffix,
        }


class TokenAttrList(tools.Serializable):
    def __init__(self, token_attrs):
        self.token_attrs = [TokenAttr(token_attr, "") for token_attr in token_attrs]

    def reset(self):
        for token_attr in self.token_attrs:
            token_attr.set("")

    @property
    def unset_token_attrs(self):
        return [token_attr for token_attr in self.token_attrs if token_attr.label == ""]

    def purge_tokens(self, input_token_attrs=None):
        """ Removes all specified token_attrs that exist in instance.token_attrs
        
        :param token_attrs: list(str), list of string values of tokens to remove.  If None, removes all
        """
        if input_token_attrs is None:
            remove_attrs = self.token_attrs
        else:
            remove_attrs = [
                token_attr
                for token_attr in self.token_attrs
                if token_attr.token in input_token_attrs
            ]

        self.token_attrs = [
            token_attr for token_attr in self.token_attrs if token_attr not in remove_attrs
        ]

    @classmethod
    def from_json(cls, json_blob):
        instance = cls(list(json_blob))
        instance.merge_json(json_blob)
        return instance

    def merge_token_attr(self, token_attr):
        if self.has_token_attr(token_attr):
            getattr(self, token_attr.token).merge_json(token_attr.to_json())
        else:
            self.token_attrs.append(token_attr)

    def has_token_attr(self, token):
        return any([token_attr for token_attr in self.token_attrs if token_attr.token == token])

    def merge_json(self, json_blob):
        for token_name, token_attr_blob in json_blob.items():
            token_name = token_name.lower()
            try:
                if not isinstance(token_attr_blob, dict):
                    token_attr_blob = {"token": token_name, "label": token_attr_blob}

                getattr(self, token_name).merge_serialization(token_attr_blob)
            except (AttributeError, IndexError):
                self.merge_token_attr(TokenAttr.from_json(token_attr_blob))

    def to_json(self):
        return {token_attr.token: token_attr.to_json() for token_attr in self.token_attrs}

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return all(
                map(
                    lambda x: x[0] == x[1],
                    zip(
                        sorted([t for t in self.token_attrs], key=lambda x: x.token),
                        sorted([t for t in other.token_attrs], key=lambda x: x.token),
                    ),
                )
            )
        return False

    def __getattr__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            try:
                return [token_attr for token_attr in self.token_attrs if token_attr.token == item][
                    0
                ]
            except IndexError:
                pass
        raise AttributeError

    def __str__(self):
        return " ".join(
            ["%s:%r" % (token_attr.token, token_attr.label) for token_attr in self.token_attrs]
        )

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.token_attrs)

    def __iter__(self):
        return iter(self.token_attrs)

    def __getitem__(self, item):
        return [token_attr for token_attr in self.token_attrs if token_attr.token == item][0]



# Is not a hard coded (word) and does not end with any non word characters or capitals (assuming camel)

DEFAULT_YML_CONFIG_FILE = 'env.yml'

SEPARATORS = r'\(\)\{\}\[\]<>.()_\-'
REGEX_SINGLE_PARENTHESIS = r'(\()|(\))'
REGEX_PARENTHESIS = r'([\(\)]+)'
REGEX_BRACKETS = r'([\{\}]+)'
REGEX_STATIC_TOKEN = r'(\(\w+\))'
REGEX_BRACKET_TOKEN = r'(\{\w+\})'
REGEX_TOKEN_SEARCH = r'(?P<token>((?<![a-z\(\)]){TOKEN}(?![0-9]))|((?<=[a-z_\(\)]){TOKEN_CAPITALIZED}(?![0-9])))'
REGEX_ADJACENT_UNDERSCORE = r'(^[\W_]+)|([\W_]+$)'
REGEX_SINGLE_LETTER = r'[a-zA-Z]'
REGEX_TOKEN = r'[A-Za-z0-9][^A-Z\W{SEP}]+'.format(SEP=SEPARATORS)

FORMAT_STRING_REGEX = r'(?:{STATIC})|({TOKEN})|{CHAR}'.format(STATIC=REGEX_STATIC_TOKEN,
                                                              TOKEN=REGEX_TOKEN,
                                                              CHAR=REGEX_SINGLE_LETTER)



#!/usr/bin/env python
from collections import Counter
import re
import string
import nomenclate.settings as settings
from . import processing


class InputRenderer(type):
    RENDER_FUNCTIONS = {}

    def __new__(mcs, name, bases, dct):
        cls = type.__new__(mcs, name, bases, dct)
        token = dct.get("token", None)
        if token:
            mcs.RENDER_FUNCTIONS[token] = cls
        return cls

    @classmethod
    def render_unique_tokens(cls, nomenclate_object, token_values):
        for token, token_settings in token_values.items():
            if token_settings.get("label") is not None and hasattr(nomenclate_object, token):
                value = token_settings.pop("label")
                token_settings.pop("token")

                renderer = cls.get_valid_render_function(token)
                if callable(getattr(renderer, "render")):
                    token_config = nomenclate_object.get_token_settings(token)
                    rendered_token = renderer.render(
                        value, token, nomenclate_object, **token_config
                    )
                    token_settings["label"] = rendered_token

    @classmethod
    def get_valid_render_function(cls, token_name):
        token_name = token_name.lower()
        renderer = None

        for func in list(cls.RENDER_FUNCTIONS):
            if token_name.replace(func, "").isdigit() or not token_name.replace(func, ""):
                renderer = func

        return cls.RENDER_FUNCTIONS.get(renderer or "default")

    @classmethod
    def render_nomenclative(cls, nomenclate_object):
        nomenclative = processing.Nomenclative(nomenclate_object.format)
        token_values = nomenclate_object.token_dict.to_json()
        cls.render_unique_tokens(nomenclate_object, token_values)
        render_template = nomenclate_object.format
        cls._prepend_token_match_objects(token_values, render_template)
        for token, match in token_values.items():
            nomenclative.add_match(*match)

        render_template = cls.cleanup_formatted_string(nomenclative.process_matches())
        return render_template

    @classmethod
    def _prepend_token_match_objects(cls, token_values, incomplete_nomenclative):
        for token, token_settings in token_values.items():
            value = token_settings["label"]
            regex_token = token.replace("(", "\(").replace(")", "\)")
            re_token = settings.REGEX_TOKEN_SEARCH.format(
                TOKEN=regex_token, TOKEN_CAPITALIZED=regex_token.capitalize()
            )
            re_matches = re.finditer(re_token, incomplete_nomenclative, 0)

            for re_match in re_matches:
                token_values[token] = (re_match, value)

        cls._clear_non_matches(token_values)

    @staticmethod
    def _clear_non_matches(token_values):
        to_delete = []
        for token, value in token_values.items():
            if isinstance(value, str) or not isinstance(value, tuple):
                to_delete.append(token)

        for delete in to_delete:
            token_values.pop(delete)

    @classmethod
    def cleanup_formatted_string(cls, formatted_string):
        """ Removes unused tokens/removes surrounding and double underscores

        :param formatted_string: str, string that has had tokens replaced
        :return: str, cleaned up name of object
        """
        # Remove whitespace
        result = formatted_string.replace(" ", "")
        # Remove any static token parentheses
        result = re.sub(settings.REGEX_PARENTHESIS, "", result)
        # Remove any multiple separator characters
        multi_character_matches = re.finditer("[%s]{2,}" % settings.SEPARATORS, result)
        for multi_character_match in sorted(
            multi_character_matches, key=lambda x: len(x.group()), reverse=True
        ):
            match = multi_character_match.group()
            most_common_separator = Counter(list(multi_character_match.group())).most_common(1)[0][
                0
            ]
            result = result.replace(match, most_common_separator)
        # Remove trailing or preceding non letter characters
        result = re.sub(settings.REGEX_ADJACENT_UNDERSCORE, "", result)
        #  not sure what this one was...but certainly not it.
        result = re.sub(settings.REGEX_SINGLE_PARENTHESIS, "", result)
        return result

    @staticmethod
    def _get_alphanumeric_index(query_string):
        """ Given an input string of either int or char, returns what index in the alphabet and case it is

        :param query_string: str, query string
        :return: (int, str), list of the index and type
        """
        # TODO: could probably rework this. it works, but it's ugly as hell.
        try:
            return [int(query_string), "int"]
        except ValueError:
            if len(query_string) == 1:
                if query_string.isupper():
                    return [string.ascii_uppercase.index(query_string), "char_hi"]
                elif query_string.islower():
                    return [string.ascii_lowercase.index(query_string), "char_lo"]
            else:
                raise IOError("The input is a string longer than one character")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nomenclate', 'nomenclate.core']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'nomenclate',
    'version': '2.5.4',
    'description': 'A tool set for automating and generating strings based on arbitrary user-defined naming conventions.',
    'long_description': 'Nomenclate: A tool set for automating and generating strings based on arbitrary user-defined naming conventions\n###############################################################################################################\n\n`Online Documentation (ReadTheDocs) <http://nomenclate.readthedocs.io/en/latest/>`_\n\n.. image:: https://readthedocs.org/projects/nomenclate/badge/?version=latest\n    :target: http://nomenclate.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n.. image:: https://badge.fury.io/py/nomenclate.svg\n    :target: https://badge.fury.io/py/nomenclate\n\n.. image:: https://circleci.com/gh/AndresMWeber/Nomenclate.svg?style=svg\n    :target: https://circleci.com/gh/AndresMWeber/Nomenclate\n\n.. image:: https://coveralls.io/repos/github/AndresMWeber/Nomenclate/badge.svg?branch=master\n    :target: https://coveralls.io/github/AndresMWeber/Nomenclate?branch=master\n\n.. image:: https://img.shields.io/pypi/pyversions/nomenclate.svg\n   :target: https://pypi.python.org/pypi/nomenclate\n\n.. contents::\n\n.. section-numbering::\n\nSynopsis\n=============\n\nNomenclate is a tool which creates persistent objects that can be used to generate strings that follow naming\nconventions that you designate.\nThere are sets of current naming conventions (format strings) that can be replaced or extended following certain rules\nfor creation. You can add arbitrary tokens as needed and register token filtering of your own designation.\n\nThere is a full set of YAML defined suffix/side substitution strings as found in ``env.yml``.\nIf you want you can create your own .yml file that you will pass to a Nomenclate instance to have your own configuration.\n\nConcept Definitions\n-------------------\ntoken\n    : A component of the format string which is a meaningful symbol/definition pair that will be filtered by\n    a grammar of regular expressions.\n    A simplified representation could be token=value wherein the token (as found in the format string) will be resolved\n    to the value as is adheres to the token\'s syntax/grammar rules\n\nformat string\n    : A string that represents a series of tokens separated with arbitrary delimiters.\n\n    e.g. - ``side_location_nameDecoratorVar_childtype_purpose_type``\n\n    Note: Nomenclate automatically supports camelCasing the tokens to separate them as a natural delimiter.\n\n`For a review of parsing/composition look here <https://en.wikipedia.org/wiki/Parsing>`_\n\nFeatures\n--------\n-  Applies a naming convention with arbitrary syntax/grammar to the formatting of string tokens\n-  Top down parsing of format string given token-specific grammar rule classes that are extensible\n-  Persistent state object instances\n-  Up to date with online help docs\n-  User-customizable YAML/human-readable config file\n-  Easy object property or dictionary state manipulation\n-  Cross-Python compatible: Tested and working with Python 3.6, 3.7 and 3.8\n-  Cross-Platform compatible: Works under Linux, Mac OS ,Windows environments\n-  Full module/class documentation\n-  Sensible token value entry/conversion (like ``side=\'left\'`` with automatic token syntax replacement)\n\nInstallation\n============\nWindows, etc.\n-------------\nA universal installation method (that works on Windows, Mac OS X, Linux, ..., and always provides the latest version) is to use `pip`:\n\n.. code-block:: bash\n\n    $ pip install Nomenclate\n\n.. code-block:: bash\n\n    $ poetry add Nomenclate\n\n\n(If ``pip`` installation fails for some reason, you can try ``easy_install nomenclate`` as a fallback.)\n\nUsage\n=============\n\nPython Package Usage\n---------------------\nUse this tool via package level functions\n\n.. code-block:: python\n\n    import nomenclate\n    # Create empty name object\n    nomenclate_empty = nomenclate.Nom()\n\n    # At any time you can query the state of the nomenclate object through the state property\n    >>> nomenclate_empty.state\n    {\'name\': \'\', \'childtype\': \'\', \'location\': \'\', \'var\': \'\', \'type\': \'\', \'side\': \'\', \'decorator\': \'\', \'purpose\': \'\'}\n\n    # You can also create a nomenclate with initialized kwargs\n    nomenclate_init_kwargs = nomenclate.Nom(name=\'test\', type=\'group\')\n\n    # Your Nomenclate object has now been initialized and all of the default token set have been added based on\n    # The default format_string property from the env.yml config file\n    # default: side_location_nameDecoratorVar_childtype_purpose_type\n    >>> nomenclate_init_kwargs.state\n    {\'name\': \'test\', \'childtype\': \'\', \'location\': \'\', \'var\': \'\', \'type\': \'group\', \'side\': \'\', \'decorator\': \'\', \'purpose\': \'\'}\n\n    # Feel free to manipulate each token\'s value on a property basis\n    >>> nomenclate_init_kwargs.location = \'rear\'\n\n    # Now that you\'re all set up you can use the get method to obtain a string representation of your conventionalized output:\n    >>> nomenclate_init_kwargs.get()\n    \'rr_test_GRP\'\n\n    # As you\'ll notice both tokens group and location have been composed following the replacements that can be found in the config YAML file.  This way things like "left" just need to be entered as "left" and then based on the yaml will replace automatically with anything you want.  Finally you don\'t need to enter things like "L" and worry about it later on!\n\n    # The format string will automate the process of hot swapping naming formats allows any string to be input.\n    >>> nomenclate_init_kwargs.format\n    \'side_location_nameDecoratorVar_childtype_purpose_type\'\n    >>> nomenclate_init_kwargs.format = \'name_type\'\n    >>> nomenclate_init_kwargs.state\n    {name:\'test\', type=\'group\'}\n\n    # You can enter static text that will always be present in the name by surrounding with parenthesis\n    # For now they only support alphanumeric characters.\n\n    >>> nomenclate_init_kwargs.format = \'side_location_nameDecoratorVar_(static.text)childtype_purpose_type\'\n    >>> nomenclate_init_kwargs.name = \'test\'\n    >>> nomenclate_init_kwargs.location = \'rear\'\n    >>> nomenclate_init_kwargs.type = \'group\'\n    \'rr_test_staticText_GRP\'\n\n    # Now entering all these values by properties is fun and all, however there is a convenience function that can digest dictionaries\n    >>> test_nom = nomenclate.Nom()\n    >>> test_nom.merge_serialization({\'name\':\'test\', \'location\':\'rear\', \'type\':\'group\'})\n    >>> test_nom.get()\n\n    # As you might have guessed, using state and merge_serialization you can pass naming values from instance to instance (as you can see __eq__ has been defined for Nomenclate instances):\n    >>> nom_a = nomenclate.Nom(name=\'test\', location=\'rear\')\n    >>> nom_b = nomenclate.Nom()\n    >>> nom_b == nom_a\n    False\n    >>> nom_b.merge_serialization(nom_a.state)\n    >>> nom_b == nom_a\n    True\n\n    # Optionally you can just pass the nomenclate object itself\n    >>> nom_b.token_dict.reset() # Internal function to be made into a public method later...\n    >>> nom_b == nom_a\n    False\n    >>> nom_b.merge_serialization(nom_a)\n    >>> nom_b == nom_a\n    True\n\n\n\nYAML Configuration File Rules\n-----------------------------\n\nSo far the suffixes is a look up dictionary for Maya objects, however I will be adding support for more later.\n\nTo properly enter a naming format string:\n\n    Enter all tokens you want to use with descriptive value that naming token\'s label e.g:\n        ``name``\n\n    and place it where you want it in order in the formatting string you set.\n    If you want something to space out or separate the names just input whatever separator\n    you want to use like ``_`` or ``.`` and it will keep those as delimiters.\n    ``name_side_type``\n\n    Additionall if you want them camel cased for example name and type:\n    ``side_nameType``\n    and it will automatically camelcase your for whatever you input for the given token values.\n\n    In the config YAML file ``(default is nomenclate/core/env.yml)`` define your format under the header ``naming_formats`` with a sub-section name you think is appropriate (the following example is optionally nested under "node"):\n\n    .. code-block:: yaml\n\n        naming_formats:\n            node:\n                your_format: name_sidePurpose_type\n\n\n    If you want a static string to always be present in a format string just enclose it with parenthesis (for now only alphanumeric characters are accepted), for example a version:\n        ``(v)version``\n        in format string:\n        ``side_name_(v)version_(static_text_example)``\n\n        Example:\n            If version is 3 and your version padding config is set to 2\n            will evaluate to:\n            ``v02``\n\nFurther version/var/date specific token notes:\n    There are 3 naming tokens with specific formatting functions that will give you customized results.  You can designate multiple fields for added granularity by adding a number after e.g. var1, var2\n\n      :var:\n        this depends on var in the config being set to upper or lower\n\n        ``a``: returns a character based on position in alphabet, if you go over it starts aa -> az -> ba -> bz etc.\n\n        ``A``: returns a character based on position in alphabet, if you go over it starts AA -> AZ -> BA -> BZ etc.\n\n      :version:\n        Will return a string number based on the version_padding config setting\n\n      :date:\n        Will return a date as a string based on a datetime module formatted string\n        that the user will input or default to YYYY-MM-DD\n\n        Please specify whichever separators (or lack of) you want to override the default behavior just modify the config\n\n        The full list of options can be found here:\n        `Datetime Documentation <https://docs.python.org/2/library/datetime.html#strftime-strptime-behavior>`_\n\n If you need any custom token value conversion functions you can specify them by inheriting from ``nomenclate.core.rendering.RenderBase`` and implementing its render function like so:\n\n    .. code-block:: python\n\n        import nomenclate\n\n        class RenderCustom(nomenclate.core.rendering.RenderBase):\n            token = \'custom\'\n            def render(cls, value, token, nomenclate_object, **kwargs):\n                """ Always prepend "meh"\n\n                :param value: str, the un-parsed/formatted token value\n                :param token: str, the name of the token in question\n                :param nomenclate_object: nomenclate.Nom, the nomenclate instance (for checking attribute values/config settings)\n                :return: str, the final syntax adhering token value\n                """\n                return \'meh\' + value\n\n    Otherwise, unless you specify an options list for a specific naming token in the custom renderer\n    it will just replace the text with whatever you set that naming token to\n    on the nomenclate object.  The options lists will be used as a filter for the\n    naming token validity or as a look up table for UIs and if you specify\n    different lengths after it. It will use the first in the list unless\n    otherwise specified in the overall_config section under "<naming_token>_length"\n    If there is no abbreviation list afterwards then just write it as a list with -\n\n\nVersion Support\n===============\nCurrently this package supports Python 2.7, 3.5 and 3.6\n\nAttribution\n===========\nWPZOOM Developer Icon Set by WPZOOM License_ Source_ - Designed by David Ferreira.\n    .. _License: http://creativecommons.org/licenses/by-sa/3.0/\n    .. _Source: http://www.wpzoom.com\n\nIcon made by iconauth_ from www.flaticon.com\n    .. _iconauth: https://www.flaticon.com/authors/freepik\n',
    'author': 'Andres Weber',
    'author_email': 'andresmweber@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://nomenclate.andresmweber.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
import builtins
import datetime
import dateutil.parser as p
from . import rendering
from . import errors as exceptions
from .tools import gen_dict_key_matches, flatten


class RenderBase(metaclass=rendering.InputRenderer):
    token = "default"

    @classmethod
    def render(
        cls,
        value,
        token,
        nomenclate_object,
        config_query_path=None,
        return_type=list,
        use_value_in_query_path=True,
        **kwargs
    ):
        """ Default renderer for a token.  It checks the config for a match, if not found it uses the value provided.

        :param value: str, value we are trying to match (or config setting for the token)
        :param token: str, token we are searching for
        :param nomenclate_object: nomenclate.core.nomenclate.Nomenclate, instance of nomenclate object to query
        :param kwargs: any config settings that relate to the token as found from the nomenclate instance
        :return: str, the resulting syntactically rendered string
        """
        if config_query_path == None:
            config_query_path = nomenclate_object.OPTIONS_PATH + [token]
            if use_value_in_query_path:
                config_query_path += [value]

        config_matches = cls.get_config_match(
            value, config_query_path, return_type, nomenclate_object, **kwargs
        )
        options = cls.flatten_input(config_matches, value)
        option = cls.process_criteria(token, options, **kwargs) if options else value
        return cls.process_token_augmentations(option, token_attr=getattr(nomenclate_object, token))

    @classmethod
    def process_token_augmentations(cls, value, token_attr):
        """ Uses any found augmentations from the TokenAttr to augment the final rendered value.  Currently
            this only processes the augmentations:
                TokenAttr().case
                TokenAttr().prefix
                TokenAttr().suffix

        :param value: str, the resulting rendered string from the TokenAttr
        :param token_attr: nomenclate.core.tokens.TokenAttr, the processed TokenAttr to be used to query settings.
        :return: str, final augmented string
        """
        return "{}{}{}".format(
            token_attr.prefix, getattr(str, token_attr.case, str)(value), token_attr.suffix
        )

    @classmethod
    def get_config_match(cls, query_string, entry_path, return_type, nomenclate_object, **kwargs):
        """ Queries the nomenclate's config data for corresponding entries and filters against the incoming
            filter_kwargs as detailed in cls.process_criteria

        :param query_string: str, string to look for once we find a match in the config
        :param token: str, the given token to search for
        :param entry_path: list(str), the query path to the subsection we are looking for to get a match in the config
        :param return_type: type, the type of return value we want (usually should be <type str>)
        :param nomenclate_object: nomenclate.core.nomenclature.Nomenclate, instance to query against (has config data)
        :return: object, whatever return type was specified
        """
        try:
            return nomenclate_object.CFG.get(entry_path, return_type=return_type)
        except exceptions.ResourceNotFoundError:
            return query_string

    @classmethod
    def flatten_input(cls, options, query_string):
        """ Takes a list, dict or str of options and outputs a filtered list
            Behavior list:
                dict: flattens/filters the dict to only key matches based on query_string
                list: just flattens a list just in case it's nested.
                str: returns a list with the string in it

        :param options: list, dict, str, input options to flatten/filter
        :param query_string: str, string we are looking for if the input is a dictionary
        :return: list, flattened list.
        """
        if not isinstance(options, (dict, list)):
            options = [options]
        else:
            if isinstance(options, dict):
                options = list(gen_dict_key_matches(query_string, options))
            options = list(flatten(options))
        return options

    @classmethod
    def process_criteria(cls, token, options, **kwargs):
        """ Each kwarg passed is considered a filter.  The kwarg is in format <token>_<filter function> and if the
            filter function is found in __builtins__ it uses the filter function and checks the result against
            the kwarg's value.  If it passes the check it is filtered out of the current list of options

        :param token: str, token we are querying
        :param options: list(str), the options to filter with kwargs
        :param kwargs: dict(str: str), dictionary of {<token>_<__builtin__ function>: compare value}
        :return:
        """
        options = list(options)
        criteria_matches = list(options)
        for criteria_function_name, criteria in kwargs.items():
            if not criteria_function_name and not criteria:
                continue
            criteria_function_name = criteria_function_name.replace("%s_" % token, "")

            try:
                builtin_func = getattr(builtins, criteria_function_name)
                criteria_matches = [
                    option for option in options if builtin_func(option) == criteria
                ]
            except AttributeError:
                pass

            if not criteria_matches:
                criteria_matches = [min(options, key=lambda x: abs(builtin_func(x) - criteria))]
        return criteria_matches[0] if criteria_matches else options[0]


class RenderDate(RenderBase):
    token = "date"

    @classmethod
    def render(
        cls,
        date,
        token,
        nomenclate_object,
        config_query_path=None,
        return_type=list,
        use_value_in_query_path=True,
        **kwargs
    ):
        if date == "now":
            d = datetime.datetime.now()
        else:
            try:
                d = p.parse(date)
            except ValueError:
                return ""
        date_format = getattr(nomenclate_object, "%s_format" % cls.token, "%Y-%m-%d")
        return d.strftime(date_format)


class RenderVar(RenderBase):
    token = "var"

    @classmethod
    def render(
        cls,
        var,
        token,
        nomenclate_object,
        config_query_path=None,
        return_type=list,
        use_value_in_query_path=True,
        **kwargs
    ):
        var_format = kwargs.get("%s_format" % cls.token, "A")
        if isinstance(var, int):
            var = cls._get_variation_id(var, var_format.isupper())
        return cls.process_token_augmentations(var, token_attr=getattr(nomenclate_object, token))

    @staticmethod
    def _get_variation_id(value, capital=False):
        """ Convert an integer value to a character. a-z then double aa-zz etc
        Args:
            value (int): integer index we're looking up
            capital (bool): whether we convert to capitals or not
        Returns (str): alphanumeric representation of the index
        """
        # Reinforcing type just in case a valid string was entered
        value = int(value)
        base_power = base_start = base_end = 0
        while value >= base_end:
            base_power += 1
            base_start = base_end
            base_end += pow(26, base_power)
        base_index = value - base_start

        # create alpha representation
        alphas = ["a"] * base_power
        for index in range(base_power - 1, -1, -1):
            alphas[index] = chr(int(97 + (base_index % 26)))
            base_index /= 26

        characters = "".join(alphas)
        return characters.upper() if capital else characters


class RenderLod(RenderVar):
    token = "lod"


class RenderVersion(RenderBase):
    token = "version"

    @classmethod
    def render(
        cls,
        version,
        token,
        nomenclate_object,
        config_query_path=None,
        return_type=list,
        use_value_in_query_path=True,
        **kwargs
    ):
        padding = kwargs.get("%s_padding" % token, 4)
        version_string = "%0{0}d"
        version = version_string.format(padding) % int(version)
        return cls.process_token_augmentations(
            version, token_attr=getattr(nomenclate_object, token)
        )


class RenderType(RenderBase):
    token = "type"

    @classmethod
    def render(
        cls,
        engine_type,
        token,
        nomenclate_object,
        config_query_path=None,
        return_type=list,
        use_value_in_query_path=True,
        **kwargs
    ):
        return super(RenderType, cls).render(
            engine_type,
            cls.token,
            nomenclate_object,
            config_query_path=nomenclate_object.SUFFIXES_PATH,
            return_type=dict,
            **kwargs
        )

